import json

from autobahn.twisted import websocket
from twisted.internet import defer, reactor


class ClientProtocol(websocket.WebSocketClientProtocol):

    def onOpen(self):
        """
        Called after the handshake has been established.
        """
        data = {
            'action': self.factory.action,
            'payload': self.factory.payload,
        }

        self.write(data)

    def onClose(self, clean, code, why):
        """
        Called when the connection has been closed.

        :param clean: True if the connection was closed cleanly.
        :type clean: bool

        :param code: The result code.
        :type code: int

        :param why: An object containing information about the closed connection.
        :type why: object
        """
        # Stop the engine.
        reactor.stop()

    def onMessage(self, data, binary):
        """
        Called evertime a new message has been recieved.

        :param data: The message recieved.
        :type data: str

        :param binary: True if the message is in binary.
        :type binary: bool
        """
        # Convert UTF8 encoded string to unicode.
        data = data.decode('utf8')

        # Deserialize JSON string to dict.
        data = json.loads(data)

        action = data.get('action')
        payload = data.get('payload')

        if action not in ['CREATE_DONE', 'UPDATE_DONE', 'DELETE_DONE', 'FETCH_DONE']:
            self.factory.deferred.errback(Exception(payload))
        else:
            if action == 'FETCH_DONE':
                if payload is None:
                    payload = 'The user does not exist.'
            elif action == 'DELETE_DONE':
                if payload:
                    payload = '{0} user deleted.'.format(payload)
                else:
                    payload = 'The user does not exist.'

            self.factory.deferred.callback(payload)

        # Close connection.
        self.sendClose()

    def write(self, payload):
        """
        Serialize payload to a JSON string and send it over the network.

        :param payload: The dict object what will be converted to JSON.
        :type payload: dict
        """
        # Serialize to JSON string.
        data = json.dumps(payload)

        # Always send a UTF8 encoded string.
        self.sendMessage(data.encode('utf8'))


class ClientProtocolFactory(websocket.WebSocketClientFactory):

    protocol = ClientProtocol

    def __init__(self, action, payload, deferred, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.deferred = deferred
        self.payload = payload
        self.action = action


def connect(command, host, port, payload):
    d = defer.Deferred()

    print('Connecting to: {0}:{1}...'.format(host, port))

    reactor.connectTCP(host, port, ClientProtocolFactory(command, payload, d))

    return d


def create_user(host, port, payload):
    return connect('CREATE', host, port, payload)


def update_user(host, port, payload):
    return connect('UPDATE', host, port, payload)


def delete_user(host, port, payload):
    return connect('DELETE', host, port, payload)


def fetch_user(host, port, payload):
    return connect('FETCH', host, port, payload)
