import json

from autobahn.twisted import websocket
from twisted.internet import threads

from .models import User


class UserCreate:

    def message(self, data):
        """
        Create new user.

        :param data: The user fields to insert to a new row.
        :type data: dict
        """
        d = threads.deferToThread(self.create, **data)

        d.addCallbacks(self.done, self.fail)

        return d

    def create(self, **kwargs):
        return User.create(**kwargs)

    def done(self, user):
        print('USER CREATED: {0}'.format(user))

        return {'action': 'CREATE_DONE', 'payload': user.dictionary}

    def fail(self, error):
        print('USER NOT CREATED: {0}'.format(error))

        return {'action': 'CREATE_FAIL', 'payload': error}


class UserUpdate:

    def message(self, data):
        """
        Update an existing user with the given primary key.

        :param data: The user primary key and fields to update.
        :type data: dict
        """
        pk = data.pop('id')

        d = threads.deferToThread(self.update, pk, data)

        d.addCallbacks(self.done, self.fail)

        return d

    def update(self, pk, fields):
        """
        Update an existing user with the given primary key.

        :param pk: The user primary key.
        :type pk: int

        :param fields: The fields to update.
        :type fields: dict
        """
        result = User.update(**fields).where(User.id == pk).execute()

        # Return the updated user info.
        if result:
            return User.select().where(User.id == pk).get().dictionary

        return result

    def done(self, result):
        print('USER UPDATED: {0}'.format(result))

        return {'action': 'UPDATE_DONE', 'payload': result}

    def fail(self, error):
        print('USER NOT UPDATED: {0}'.format(error))

        return {'action': 'UPDATE_FAIL', 'payload': error}


class UserDelete:

    def message(self, data):
        """
        Delete an existing user with the given primary key.

        :param data: The user primary key.
        :type data: int
        """
        d = threads.deferToThread(self.action, data)

        d.addCallbacks(self.done, self.fail)

        return d

    def action(self, pk):
        """
        Delete an existing user with the given primary key.

        :param pk: The user primary key.
        :type pk: int
        """
        return User.delete().where(User.id == pk).execute()

    def done(self, n):
        print('USER DELETED: {0}'.format(n))

        return {'action': 'DELETE_DONE', 'payload': n}

    def fail(self, error):
        print('USER NOT DELETED: {0}'.format(error))

        return {'action': 'DELETE_FAIL', 'payload': error}


class UserFetch:

    def message(self, data):
        """
        Fetch the user with the given primary key. If the primary key
        is not provided, fetch all users.

        :param data: The user primary key.
        :type data: int
        """
        d = threads.deferToThread(self.fetch, data)

        d.addCallbacks(self.done, self.fail)

        return d

    def fetch(self, pk):
        if pk:
            try:
                return User.get(User.id == pk).dictionary
            except User.DoesNotExist:
                return None

        return [user.dictionary for user in User.select()]

    def done(self, result):
        print('USER(S) SELECTED: {0}'.format(result))

        return {'action': 'FETCH_DONE', 'payload': result}

    def fail(self, error):
        print('USER SELECT ERROR: {0}'.format(error))

        return {'action': 'FETCH_FAIL', 'payload': error}


class ServerProtocol(websocket.WebSocketServerProtocol):

    def onMessage(self, data, binary):
        """
        Called evertime a new message has been recieved.

        :param data: The message recieved.
        :type data: str

        :param binary: True if the message is in binary.
        :type binary: bool
        """
        # Convert UTF8 encoded string to unicode.
        data = data.decode('utf8')

        # Deserialize JSON string to dict.
        data = json.loads(data)

        action, payload = (data.get('action'),
                           data.get('payload'))

        try:
            # Get action handler.
            handler = self.factory.actions[action]

            d = handler.message(payload)

            d.addBoth(self.write)
        except KeyError:
            self.write({'action': 'ACTION_FAIL', 'payload': 'Invalid action'})

    def write(self, payload):
        """
        Serialize payload to a JSON string and send it over the network.

        :param payload: The dict object what will be converted to JSON.
        :type payload: dict
        """
        # Serialize to JSON string.
        data = json.dumps(payload)

        # Always send a UTF8 encoded string.
        self.sendMessage(data.encode('utf8'))


class ServerProtocolFactory(websocket.WebSocketServerFactory):

    protocol = ServerProtocol

    actions = {
        'CREATE': UserCreate(),
        'UPDATE': UserUpdate(),
        'DELETE': UserDelete(),
        'FETCH': UserFetch(),
    }
